#pragma implementation

#include "DebtagsEnvironment.h"

#include <StdioParserInput.h>
#include <TagcollParser.h>
#include <TagcollFilter.h>
#include <DerivedTags.h>
#include <FilterChain.h>
#include <Implications.h>
#include <Patches.h>
#include <TagcollSerializer.h>

#include "config.h"

#include <sys/types.h>	// getpwuid, stat, mkdir, getuid
#include <sys/stat.h>	// stat, mkdir
#include <pwd.h>		// getpwuid
#include <unistd.h>		// stat, getuid

#include <errno.h>

using namespace std;
using namespace Debtags;

/*
extern const char* fn_tagdb;
extern const char* fn_impls;
extern const char* fn_deriv;
*/
const char* fn_tagdb = "/var/lib/debtags/package-tags";
const char* fn_impls = "/var/lib/debtags/implications";
const char* fn_deriv = "/var/lib/debtags/derived-tags";
const char* fn_vocab = "/var/lib/debtags/vocabulary";
const char* fn_dist_vocab = "/usr/share/debtags/vocabulary";
const char* fn_sources = "/etc/debtags/sources.list";
const char* path_cache = "/var/cache/debtags";
const char* path_tagpatch_d = "/etc/debtags/tagpatch.d";	
const char* path_tagvoc_d = "/etc/debtags/tagvoc.d";	
const char* remote_tags_fname = "tags-current.gz";

static void readCollection(const string& file, TagcollConsumer<string>& consumer)
	throw (FileException, ParserException)
{
	StdioParserInput in(file);
	TagcollParser::parseTagcoll(in, consumer);
}

static void readDerivedTags(const string& file, DerivedTagList& derivedTags)
	throw (FileException)
{
	if (file == "-")
	{
		StdioParserInput input(stdin, "<stdin>");
		derivedTags.parse(input);
	}
	else
	{
		StdioParserInput input(file);
		derivedTags.parse(input);
	}
}

namespace Debtags
{

// TagcollFilter that removes that which are not faceted
template<class ITEM, class TAG = std::string>
class UnfacetedRemover : public TagcollFilter<ITEM, TAG>
{
public:
	virtual void consume(const std::string& item) throw ();
	virtual void consume(const std::string& item, const OpSet<std::string>& tags) throw ();
	virtual void consume(const OpSet<ITEM>& items) throw ();
	virtual void consume(const OpSet<ITEM>& items, const OpSet<TAG>& tags) throw ();
};

};

template<class ITEM, class TAG>
void UnfacetedRemover<ITEM, TAG>::consume(const string& item) throw ()
{
	consumer->consume(item);
}

template<class ITEM, class TAG>
void UnfacetedRemover<ITEM, TAG>::consume(const string& item, const OpSet<string>& tags) throw ()
{
	OpSet<string> patched;
	for (OpSet<string>::const_iterator i = tags.begin();
			i != tags.end(); i++)
		if (i->find("::") != string::npos)
			patched += *i;

	if (patched.size())
		consumer->consume(item, patched);
	else
		consumer->consume(item);
}

// Process a set of items, all with no tags
template<class ITEM, class TAG>
void UnfacetedRemover<ITEM, TAG>::consume(const OpSet<ITEM>& items) throw ()
{
	consumer->consume(items);
}

// Process a set of items identically tagged, with their tags
template<class ITEM, class TAG>
void UnfacetedRemover<ITEM, TAG>::consume(const OpSet<ITEM>& items, const OpSet<TAG>& tags) throw ()
{
	OpSet<string> patched;
	for (OpSet<string>::const_iterator i = tags.begin();
			i != tags.end(); i++)
		if (i->find("::") != string::npos)
			patched += *i;

	if (patched.size())
		consumer->consume(items, patched);
	else
		consumer->consume(items);
}



Environment* Environment::instance = 0;
                                                                                                        
void Environment::init(bool editable) throw (SystemException, ConsistencyCheckException)
{
	if (instance == 0)
		instance = new Environment(editable);
}

Environment::Environment(bool editable) throw (SystemException, ConsistencyCheckException)
{
	struct passwd* udata = getpwuid(getuid());
	rcdir = udata->pw_dir;
	rcdir += "/.debtags";

	printf("Init debtags environment; rcdir is %.*s\n", PFSTR(rcdir));

	struct stat rcdir_stat;
	if (stat(rcdir.c_str(), &rcdir_stat) == -1)
	{
		printf("Creating profile directory %.*s.\n", PFSTR(rcdir));
		if (editable && mkdir(rcdir.c_str(), 0777) == -1)
			throw SystemException(errno, "creating directory " + rcdir);
	} else {
		if (S_ISDIR(rcdir_stat.st_mode) == 0)
			if (editable)
				throw ConsistencyCheckException(rcdir + " already exists and is not a directory");
			else
				fprintf(stderr, "warning: %.*s already exists and is not a directory", PFSTR(rcdir));
		
		if (access(rcdir.c_str(), R_OK | W_OK | X_OK) == -1)
			if (editable)
				throw ConsistencyCheckException(rcdir + " already exists and has insufficient permissions");
			else
				fprintf(stderr, "%.*s already exists and has insufficient permissions", PFSTR(rcdir));
	}
}

bool Environment::hasTagDatabase() throw ()
{
	if (   access(fn_tagdb, R_OK) == -1
		|| access(fn_impls, R_OK) == -1
		|| access(fn_deriv, R_OK) == -1
		|| access(fn_vocab, R_OK) == -1)
		return false;
	return true;
}

// Output the current Debian tags database to a TagcollConsumer
void Environment::outputSystem(TagcollConsumer<std::string, std::string>& cons) const
	throw (FileException, ParserException)
{
	// Prepare the input filter chain
	FilterChain<string> filters;

	// Expand implications
	ImplicationList implications;
	readCollection(fn_impls, implications);
	// Pack the structure for faster expansions
	implications.pack();
	TagcollFilter<string>* ximpl = new ApplyImplications(implications);
	filters.appendFilter(ximpl);

	// Add derived tags
	DerivedTagList derivedTags;
	readDerivedTags(fn_deriv, derivedTags);
	TagcollFilter<string>* xderv = new ApplyDerivedTags(derivedTags);
	filters.appendFilter(xderv);

	// Add further tags implicated by the derived tags
	filters.appendFilter(ximpl);

	// Remove unfaceted tags
	UnfacetedRemover<string> ufr;
	filters.appendFilter(&ufr);

	// Read and expand the database
	filters.setConsumer(&cons);
	readCollection(fn_tagdb, filters);

	delete ximpl;
	delete xderv;
}

// Output the current Debian tags database, patched with local patch, to a TagcollConsumer
void Environment::outputPatched(TagcollConsumer<std::string, std::string>& cons) const
	throw (FileException, ParserException)
{
	// Prepare the input filter chain
	FilterChain<string> filters;

	// Apply local patch if it exists
	PatchList patches;
	string patchFile = rcdir + "/patch";
	if (access(patchFile.c_str(), R_OK) == 0)
	{
		readCollection(patchFile, patches);
		filters.appendFilter(new ApplyPatches(patches));
	}

	// Expand implications
	ImplicationList implications;
	readCollection(fn_impls, implications);
	// Pack the structure for faster expansions
	implications.pack();
	TagcollFilter<string>* ximpl = new ApplyImplications(implications);
	filters.appendFilter(ximpl);

	// Add derived tags
	DerivedTagList derivedTags;
	readDerivedTags(fn_deriv, derivedTags);
	TagcollFilter<string>* xderv = new ApplyDerivedTags(derivedTags);
	filters.appendFilter(xderv);

	// Add further tags implicated by the derived tags
	filters.appendFilter(ximpl);

	// Remove unfaceted tags
	UnfacetedRemover<string> ufr;
	filters.appendFilter(&ufr);

	// Read and expand the database
	filters.setConsumer(&cons);
	readCollection(fn_tagdb, filters);

	delete ximpl;
	delete xderv;
}

void Environment::savePatch(const InputMerger<std::string>& collection)
	throw (FileException, ParserException, SystemException)
{
	InputMerger<string> system;
	outputSystem(system);

	PatchList patch;
	patch.generate(system, collection);

	string patchFile = rcdir + "/patch";
	string backup = patchFile + "~";

	if (access(patchFile.c_str(), F_OK) == 0)
		if (rename(patchFile.c_str(), backup.c_str()) == -1)
			throw SystemException(errno, "Can't rename " + patchFile + " to " + backup);

	try {
		FILE* out = fopen(patchFile.c_str(), "w");
		if (out == 0)
			throw SystemException(errno, "Can't write to " + patchFile);
		
		TagcollSerializer output(out);
		patch.output(output);

		fclose(out);
	} catch (Exception& e) {
		if (rename(backup.c_str(), patchFile.c_str()) == -1)
			fprintf(stderr, "Warning: Impossible to restore previous backup copy: %.*s (%s)\n", PFSTR(e.desc()), e.type());
		throw;
	}
}

void Environment::sendPatch(const InputMerger<std::string>& collection)
	throw (FileException, ParserException, SystemException, ConsistencyCheckException)
{
	InputMerger<string> system;
	outputSystem(system);

	PatchList patch;
	patch.generate(system, collection);

	static const char* cmd = "/usr/sbin/sendmail -t";
	FILE* out = popen(cmd, "w");
	if (out == 0)
		throw SystemException(errno, string("trying to run `") + cmd + "'");

	struct passwd* udata = getpwuid(getuid());

	fprintf(out,
			"To: debtags-submit@vitavonni.de\n"
			"Bcc: %s\n"
			"Subject: Tag patch\n"
			"Mime-Version: 1.0\n"
			"Content-Type: multipart/mixed; boundary=\"9amGYk9869ThD9tj\"\n"
			"Content-Disposition: inline\n"
			"X-Mailer: debtags-edit\n\n"
			"This mail contains a Debtags patch for the central archive\n\n"
			"--9amGYk9869ThD9tj\n"
			"Content-Type: text/plain; charset=utf-8\n"
			"Content-Disposition: inline\n\n"
			"-- DEBTAGS DIFF V0.1 --\n", udata->pw_name);

	TagcollSerializer output(out);
	patch.output(output);

	fprintf(out, "\n--9amGYk9869ThD9tj\n");

	int res = pclose(out);
	if (!WIFEXITED(res) || WEXITSTATUS(res) != 0)
		throw ConsistencyCheckException("sendmail returned nonzero (" + stringf::fmt(res) + "): the mail may have not been sent");
}

// vim:set ts=4 sw=4:
