/*
 * Central class to perform package searches
 *
 * Copyright (C) 2003  Enrico Zini <enrico@debian.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#pragma implementation

#include "DebtagsPackage.h"
#include "DebtagsEnvironment.h"
#include "DebDBParser.h"
#include "PackageDB.h"

#include <StdioParserInput.h>

using namespace std;
using namespace Debtags;

namespace Debtags
{
	
class AvailableReader : public DebDBParser
{
protected:
	ParserInput* input;
	Packages& pkgs;
	map<string, Packages::Maintainer*> maints;

public:
	AvailableReader(Packages& pkgs) throw () :
		DebDBParser(*(input = new StdioParserInput("/var/lib/dpkg/available"))),
		pkgs(pkgs) {}

	~AvailableReader() throw ()
	{
		if (input != 0)
			delete input;
	}

	Package* nextRecord() throw (ParserException);

	void saveMaints() throw ();
};

class ItemsToPackages : public TagcollConsumer<string, string>
{
protected:
	const Packages& target;
	
public:
	ItemsToPackages(const Packages& target) throw ()
		: target(target) {}

	virtual void consume(const string& item) throw ()
	{
		target._tagdb.consume(target.obtain(item));
	}
	virtual void consume(const string& item, const OpSet<string>& tags) throw ()
	{
		target._tagdb.consume(target.obtain(item), tags);
	}
};

class PackagesToItems : public TagcollConsumer<Debtags::Package*, string>
{
protected:
	InputMerger<string>& target;

public:
	PackagesToItems(InputMerger<string>& target) throw ()
		: target(target) {}

	void consume(Debtags::Package* const& item) throw ()
	{
		target.consume(item->name());
	}
	void consume(Debtags::Package* const& item, const OpSet<string>& tags) throw ()
	{
		target.consume(item->name(), tags);
	}
};

};

// Read a record and positions itself at the start of the next one
// Returns false when there are no more records available
Package* AvailableReader::nextRecord() throw (ParserException)
{
	Package* res = 0;

	if (isEOF)
		return res;

	int n;
	do {
		// Read the field name
		string field = getFieldName();
		string body;

		//fprintf(stderr, "Got field: %.*s\n", field.size(), field.data());

		// Read the colon
		eatFieldSep();

		// Read the first line of the field body
		appendFieldBody(body);
		//fprintf(stderr, "Got body: %.*s\n", body.size(), body.data());

		// Read the continuation lines of field body
		while ((n = eatSpacesAndEmptyLines()) == 0 && ! isBOL)
		{
			body += '\n';

			size_t start_size = body.size();

			appendFieldBody(body);

			// Check for dot-only lines to be changed to empty lines
			if (body.size() - start_size == 1 && body[body.size() - 1] == '.')
				body.resize(body.size() - 1);
	
			//fprintf(stderr, "Appended body: %.*s\n", body.size(), body.data());
		}
		//fprintf(stderr, "Trailing newlines: %d\n", n);
		if (field == "Package")
		{
			res = pkgs.obtain(body);
		}
		else if (res && field == "Maintainer")
		{
			map<string, Packages::Maintainer*>::iterator i = maints.find(body);
			Packages::Maintainer* maint;
			if (i == maints.end())
			{
				maints.insert(pair<string, Packages::Maintainer*>(body, maint = new Packages::Maintainer(body, res)));
			} else {
				maint = i->second;
				maint->addPkg(res);
			}
			res->maint(maint);
		}
		else if (field == "Description")
		{
			int send = body.find('\n');
			res->desc(body.substr(0, send));
			res->ldesc(body.substr(send + 1));
		}
	} while (!isEOF && !n);

	if (isEOF)
	{
		delete input;
		input = 0;
	}

	return res;
}

void AvailableReader::saveMaints() throw ()
{
	for (map<string, Packages::Maintainer*>::const_iterator i = maints.begin();
			i != maints.end(); i++)
		pkgs._maintainers.push_back(i->second);
}

Packages::Packages() throw ()
	: availableLoaded(false), debtagsLoaded(false)
{
}

Packages::~Packages() throw ()
{
	for (map<std::string, Package*>::iterator i = pool.begin(); i != pool.end(); i++)
		delete i->second;
	for (std::vector<Maintainer*>::iterator i = _maintainers.begin();
			i != _maintainers.end(); i++)
		delete *i;
}

TagcollChange<Package*, std::string> Packages::applyChange(const TagcollChange<Package*, std::string>& change) throw ()
{
	return _tagdb.applyChange(change);
}

OpSet<std::string> Packages::getAllTags() const throw ()
{
	if (!debtagsLoaded)
		loadDebtags();
	OpSet<std::string> allTags = _tagdb.getAllTags();
	for (Vocabulary::const_iterator i = _vocab.begin(); i != _vocab.end(); i++)
		allTags += i->first;

	return allTags;
}

OpSet<std::string> Packages::getCompanionTags(const OpSet<std::string>& ts) const throw ()
{
	if (!debtagsLoaded)
		loadDebtags();
	// Does not need to look at the vocabulary: companion tags can only be
	// tags that are being used
	return _tagdb.getCompanionTags(ts);
}

OpSet<std::string> Packages::getTagsetForItem(Package * const pkg) const throw ()
{
	if (!debtagsLoaded)
		loadDebtags();
	return _tagdb.getTagsetForItem(pkg);
}

Package* Packages::obtain(const std::string& name) const throw ()
{
	map<std::string, Package*>::iterator i = pool.find(name);
	if (i != pool.end())
		return i->second;
	Package* res = new Package(*this, name);
	pool.insert(pair<string, Package*>(name, res));
	return res;
}

void Packages::loadAvailable() throw ()
{
	try {
		AvailableReader reader(*this);
		while (reader.nextRecord())
			;
		reader.saveMaints();
	} catch (Exception& e) {
		fprintf(stderr, "%s: %.*s.  Aborting loading of available file\n", e.type(), PFSTR(e.desc()));
	}
	availableLoaded = true;
}

void Packages::loadDebtags() const throw ()
{
	fprintf(stderr, "Beginning to load debtags...\n");
	_vocab.read();
	ItemsToPackages itp(*this);
	Environment::get().outputPatched(itp);
	debtagsLoaded = true;
	fprintf(stderr, "Debtags loaded.\n");
}

void Packages::sendIfMatch(Package* pkg, PackageConsumer& cons, Filter& filter) throw ()
{
	bool got = true;
	if (filter.hasMaintainer() && pkg->maint() && !filter.matchMaintainer(pkg->maint()->name()))
		got = false;
	if (filter.hasName() && !filter.matchName(pkg->name()))
		got = false;
	if (filter.hasString() && !filter.matchString(pkg->name()))
		got = false;
	if (filter.hasString() && !filter.matchString(pkg->desc()))
		got = false;
	if (filter.hasString() && !filter.matchString(pkg->ldesc()))
		got = false;
	if (filter.hasInstalled() && !filter.matchInstalled(_packagedb.getState(pkg->name())))
		got = false;
	if (filter.hasDebtags() && !filter.matchDebtags(_tagdb.getTagsetForItem(pkg)))
		got = false;
	if (got)
		cons.consume(pkg);
}


Package* Packages::find(const std::string& name) throw ()
{
	if (!availableLoaded)
		loadAvailable();
	map<string, Package*>::iterator found = pool.find(name);
	return found == pool.end() ? 0 : found->second;
}

void Packages::output(PackageConsumer& cons) throw ()
{
	if (availableLoaded)
	{
		for (map<std::string, Package*>::iterator i = pool.begin(); i != pool.end(); i++)
			cons.consume(i->second);
	} else {
		try {
			AvailableReader reader(*this);
			while (Package* pkg = reader.nextRecord())
				cons.consume(pkg);
			reader.saveMaints();
		} catch (Exception& e) {
			fprintf(stderr, "%s: %.*s.  Aborting loading of available file\n", e.type(), PFSTR(e.desc()));
		}
		availableLoaded = true;
	}
}

void Packages::output(PackageConsumer& cons, Filter& filter) throw ()
{
	if (filter.hasDebtags() && !debtagsLoaded)
		loadDebtags();
	if (filter.hasNonDebtags() && !availableLoaded)
		loadAvailable();

	if (filter.hasOnlyMaintainer() && filter.isMaintainerSignificant())
	{
		fprintf(stderr, "Strategy0\n");
		for (vector<Maintainer*>::const_iterator i = _maintainers.begin();
				i != _maintainers.end(); i++)
			if (filter.matchMaintainer((*i)->name()))
				for (vector<Package*>::const_iterator j = (*i)->packages().begin();
						j != (*i)->packages().end(); j++)
					cons.consume(*j);
	}
	else if (filter.isMaintainerSignificant())
	{
		fprintf(stderr, "Strategy1\n");
		for (vector<Maintainer*>::const_iterator i = _maintainers.begin();
				i != _maintainers.end(); i++)
			if (filter.matchMaintainer((*i)->name()))
				for (vector<Package*>::const_iterator j = (*i)->packages().begin();
						j != (*i)->packages().end(); j++)
					sendIfMatch(*j, cons, filter);
	}
	else if (filter.hasOnlyName())
	{
		fprintf(stderr, "Strategy2\n");
		for (pool_t::const_iterator i = pool.begin();
				i != pool.end(); i++)
			if (filter.matchName(i->second->name()))
				cons.consume(i->second);
	}
	else if (filter.hasName())
	{
		fprintf(stderr, "Strategy3\n");
		for (pool_t::const_iterator i = pool.begin();
				i != pool.end(); i++)
			if (filter.matchName(i->second->name()))
				sendIfMatch(i->second, cons, filter);
	}
	else if (filter.hasOnlyDebtags())
	{
		fprintf(stderr, "Strategy4\n");
		OpSet<Package*> start = _tagdb.getItemsContaining(filter.debtagsTagset());
		for (OpSet<Package*>::const_iterator i = start.begin();
				i != start.end(); i++)
			cons.consume(*i);
	}
	else if (filter.hasDebtags())
	{
		fprintf(stderr, "Strategy5\n");
		OpSet<Package*> start = _tagdb.getItemsContaining(filter.debtagsTagset());
		for (OpSet<Package*>::const_iterator i = start.begin();
				i != start.end(); i++)
			sendIfMatch(*i, cons, filter);
	}
	else
	{
		fprintf(stderr, "Strategy6\n");
		// If no special case apply, perform a generic search
		for (pool_t::const_iterator i = pool.begin();
				i != pool.end(); i++)
			sendIfMatch(i->second, cons, filter);
	}
}


void Packages::outputRelated(PackageConsumer& cons, Package* pivot, int distance) throw ()
{
	if (!debtagsLoaded)
		loadDebtags();
	OpSet<Package*> res = _tagdb.getRelatedItems(pivot, distance);
	fprintf(stderr, "%d related to %.*s, distance %d\n", res.size(), PFSTR(pivot->name()), distance);
	for (OpSet<Package*>::const_iterator i = res.begin(); i != res.end(); i++)
		cons.consume(*i);
}

void Packages::save() throw (FileException, ParserException, SystemException)
{
	if (!debtagsLoaded)
		loadDebtags();
	InputMerger<string> toSave;
	PackagesToItems pta(toSave);
	_tagdb.output(pta);
	
	Debtags::Environment::get().savePatch(toSave);
}

void Packages::send() throw (FileException, ParserException, SystemException, ConsistencyCheckException)
{
	if (!debtagsLoaded)
		loadDebtags();
	InputMerger<string> toSave;
	PackagesToItems pta(toSave);
	_tagdb.output(pta);
	
	Debtags::Environment::get().sendPatch(toSave);
}


// vim:set ts=4 sw=4:
